package ldutil

import (
	"bytes"
	"context"
	"testing"

	"github.com/stretchr/testify/require"
	blocks "gitlab.dms3.io/dms3/go-block-format"
	ld "gitlab.dms3.io/ld/go-ld-prime"
	cidlink "gitlab.dms3.io/ld/go-ld-prime/linking/cid"
	basicnode "gitlab.dms3.io/ld/go-ld-prime/node/basic"
	"gitlab.dms3.io/ld/go-ld-prime/traversal"
	"gitlab.dms3.io/ld/go-ld-prime/traversal/selector"
	"gitlab.dms3.io/ld/go-ld-prime/traversal/selector/builder"

	"gitlab.dms3.io/dms3/go-graphsync"
	"gitlab.dms3.io/dms3/go-graphsync/testutil"
)

func TestTraverser(t *testing.T) {
	ctx := context.Background()

	t.Run("traverses correctly, simple struct", func(t *testing.T) {
		testdata := testutil.NewTestLDTree()
		ssb := builder.NewSelectorSpecBuilder(basicnode.Prototype.Any)
		sel := ssb.ExploreRecursive(selector.RecursionLimitNone(), ssb.ExploreAll(ssb.ExploreRecursiveEdge())).Node()
		traverser := TraversalBuilder{
			Root:     testdata.RootNodeLnk,
			Selector: sel,
		}.Start(ctx)
		checkTraverseSequence(ctx, t, traverser, []blocks.Block{
			testdata.RootBlock,
			testdata.LeafAlphaBlock,
			testdata.MiddleMapBlock,
			testdata.LeafAlphaBlock,
			testdata.MiddleListBlock,
			testdata.LeafAlphaBlock,
			testdata.LeafAlphaBlock,
			testdata.LeafBetaBlock,
			testdata.LeafAlphaBlock,
		})
	})

	t.Run("traverses correctly, blockchain", func(t *testing.T) {
		// t.Skip()
		store := make(map[ld.Link][]byte)
		persistence := testutil.NewTestStore(store)
		blockChain := testutil.SetupBlockChain(ctx, t, persistence, 100, 10)
		inProgressChan := make(chan graphsync.ResponseProgress)
		done := make(chan struct{})
		traverser := TraversalBuilder{
			Root:       blockChain.TipLink,
			Selector:   blockChain.Selector(),
			Chooser:    blockChain.Chooser,
			LinkSystem: persistence,
			Visitor: func(tp traversal.Progress, node ld.Node, r traversal.VisitReason) error {
				select {
				case <-ctx.Done():
				case inProgressChan <- graphsync.ResponseProgress{
					Node:      node,
					Path:      tp.Path,
					LastBlock: tp.LastBlock,
				}:
				}
				return nil
			},
		}.Start(ctx)
		go func() {
			blockChain.VerifyWholeChainWithTypes(ctx, inProgressChan)
			close(done)
		}()
		checkTraverseSequence(ctx, t, traverser, blockChain.AllBlocks())
		close(inProgressChan)
		testutil.AssertDoesReceive(ctx, t, done, "should have completed verification but did not")
	})
}

func checkTraverseSequence(ctx context.Context, t *testing.T, traverser Traverser, expectedBlks []blocks.Block) {
	for _, blk := range expectedBlks {
		isComplete, err := traverser.IsComplete()
		require.False(t, isComplete)
		require.NoError(t, err)
		lnk, _ := traverser.CurrentRequest()
		// fmt.Printf("i = %v, prefix lnk = %x, blk = %x \n", i, lnk.(cidlink.Link).Cid.Prefix(), (blk.Cid()).Prefix())
		// fmt.Printf("i = %v, bytes lnk = %x, blk = %x \n", i, lnk.(cidlink.Link).Cid.Bytes(), (blk.Cid()).Bytes())
		require.Equal(t, lnk.(cidlink.Link).Cid, blk.Cid())
		err = traverser.Advance(bytes.NewBuffer(blk.RawData()))
		require.NoError(t, err)
	}
	isComplete, err := traverser.IsComplete()
	require.True(t, isComplete)
	require.NoError(t, err)
}
