package network

import (
	"context"
	"fmt"
	"io"
	"time"

	logging "gitlab.dms3.io/dms3/go-log"
	ma "gitlab.dms3.io/mf/go-multiaddr"
	"gitlab.dms3.io/p2p/go-msgio"
	"gitlab.dms3.io/p2p/go-p2p-core/host"
	"gitlab.dms3.io/p2p/go-p2p-core/network"
	"gitlab.dms3.io/p2p/go-p2p-core/peer"

	gsmsg "gitlab.dms3.io/dms3/go-graphsync/message"
)

var log = logging.Logger("graphsync_network")

var sendMessageTimeout = time.Minute * 10

// NewFromP2pHost returns a GraphSyncNetwork supported by underlying P2p host.
func NewFromP2pHost(host host.Host) GraphSyncNetwork {
	graphSyncNetwork := p2pGraphSyncNetwork{
		host: host,
	}

	return &graphSyncNetwork
}

// p2pGraphSyncNetwork transforms the p2p host interface, which sends and receives
// NetMessage objects, into the graphsync network interface.
type p2pGraphSyncNetwork struct {
	host host.Host
	// inbound messages from the network are forwarded to the receiver
	receiver Receiver
}

type streamMessageSender struct {
	s network.Stream
}

func (s *streamMessageSender) Close() error {
	return s.s.Close()
}

func (s *streamMessageSender) Reset() error {
	return s.s.Reset()
}

func (s *streamMessageSender) SendMsg(ctx context.Context, msg gsmsg.GraphSyncMessage) error {
	return msgToStream(ctx, s.s, msg)
}

func msgToStream(ctx context.Context, s network.Stream, msg gsmsg.GraphSyncMessage) error {
	log.Debugf("Outgoing message with %d requests, %d responses, and %d blocks",
		len(msg.Requests()), len(msg.Responses()), len(msg.Blocks()))

	deadline := time.Now().Add(sendMessageTimeout)
	if dl, ok := ctx.Deadline(); ok {
		deadline = dl
	}
	if err := s.SetWriteDeadline(deadline); err != nil {
		log.Warnf("error setting deadline: %s", err)
	}

	switch s.Protocol() {
	case ProtocolGraphsync:
		if err := msg.ToNet(s); err != nil {
			log.Debugf("error: %s", err)
			return err
		}
	default:
		return fmt.Errorf("unrecognized protocol on remote: %s", s.Protocol())
	}

	if err := s.SetWriteDeadline(time.Time{}); err != nil {
		log.Warnf("error resetting deadline: %s", err)
	}
	return nil
}

func (gsnet *p2pGraphSyncNetwork) NewMessageSender(ctx context.Context, p peer.ID) (MessageSender, error) {
	s, err := gsnet.newStreamToPeer(ctx, p)
	if err != nil {
		return nil, err
	}

	return &streamMessageSender{s: s}, nil
}

func (gsnet *p2pGraphSyncNetwork) newStreamToPeer(ctx context.Context, p peer.ID) (network.Stream, error) {
	return gsnet.host.NewStream(ctx, p, ProtocolGraphsync)
}

func (gsnet *p2pGraphSyncNetwork) SendMessage(
	ctx context.Context,
	p peer.ID,
	outgoing gsmsg.GraphSyncMessage) error {

	s, err := gsnet.newStreamToPeer(ctx, p)
	if err != nil {
		return err
	}

	if err = msgToStream(ctx, s, outgoing); err != nil {
		_ = s.Reset()
		return err
	}

	return s.Close()
}

func (gsnet *p2pGraphSyncNetwork) SetDelegate(r Receiver) {
	gsnet.receiver = r
	gsnet.host.SetStreamHandler(ProtocolGraphsync, gsnet.handleNewStream)
	gsnet.host.Network().Notify((*p2pGraphSyncNotifee)(gsnet))
}

func (gsnet *p2pGraphSyncNetwork) ConnectTo(ctx context.Context, p peer.ID) error {
	return gsnet.host.Connect(ctx, peer.AddrInfo{ID: p})
}

// handleNewStream receives a new stream from the network.
func (gsnet *p2pGraphSyncNetwork) handleNewStream(s network.Stream) {
	defer s.Close()

	if gsnet.receiver == nil {
		_ = s.Reset()
		return
	}

	reader := msgio.NewVarintReaderSize(s, network.MessageSizeMax)
	for {
		received, err := gsmsg.FromMsgReader(reader)
		p := s.Conn().RemotePeer()

		if err != nil {
			if err != io.EOF {
				_ = s.Reset()
				go gsnet.receiver.ReceiveError(p, err)
				log.Debugf("graphsync net handleNewStream from %s error: %s", s.Conn().RemotePeer(), err)
			}
			return
		}

		ctx := context.Background()
		log.Debugf("graphsync net handleNewStream from %s", s.Conn().RemotePeer())
		gsnet.receiver.ReceiveMessage(ctx, p, received)
	}
}

type p2pGraphSyncNotifee p2pGraphSyncNetwork

func (nn *p2pGraphSyncNotifee) p2pGraphSyncNetwork() *p2pGraphSyncNetwork {
	return (*p2pGraphSyncNetwork)(nn)
}

func (nn *p2pGraphSyncNotifee) Connected(n network.Network, v network.Conn) {
	nn.p2pGraphSyncNetwork().receiver.Connected(v.RemotePeer())
}

func (nn *p2pGraphSyncNotifee) Disconnected(n network.Network, v network.Conn) {
	nn.p2pGraphSyncNetwork().receiver.Disconnected(v.RemotePeer())
}

func (nn *p2pGraphSyncNotifee) OpenedStream(n network.Network, v network.Stream) {}
func (nn *p2pGraphSyncNotifee) ClosedStream(n network.Network, v network.Stream) {}
func (nn *p2pGraphSyncNotifee) Listen(n network.Network, a ma.Multiaddr)         {}
func (nn *p2pGraphSyncNotifee) ListenClose(n network.Network, a ma.Multiaddr)    {}
