package testutil

import (
	blocks "gitlab.dms3.io/dms3/go-block-format"
	"gitlab.dms3.io/dms3/go-cid"
	"gitlab.dms3.io/ld/go-ld-prime"

	// to register multicodec
	_ "gitlab.dms3.io/ld/go-ld-prime/codec/dagjson"
	"gitlab.dms3.io/ld/go-ld-prime/fluent"
	cidlink "gitlab.dms3.io/ld/go-ld-prime/linking/cid"
	basicnode "gitlab.dms3.io/ld/go-ld-prime/node/basic"
)

// TestLDTree is a set of LD Data that forms a tree spread across some blocks
// with a serialized in memory representation
type TestLDTree struct {
	Storage           map[ld.Link][]byte
	LeafAlpha         ld.Node
	LeafAlphaLnk      ld.Link
	LeafAlphaBlock    blocks.Block
	LeafBeta          ld.Node
	LeafBetaLnk       ld.Link
	LeafBetaBlock     blocks.Block
	MiddleMapNode     ld.Node
	MiddleMapNodeLnk  ld.Link
	MiddleMapBlock    blocks.Block
	MiddleListNode    ld.Node
	MiddleListNodeLnk ld.Link
	MiddleListBlock   blocks.Block
	RootNode          ld.Node
	RootNodeLnk       ld.Link
	RootBlock         blocks.Block
}

// NewTestLDTree returns a fake tree of nodes, spread across 5 blocks
func NewTestLDTree() TestLDTree {
	var storage = make(map[ld.Link][]byte)
	persistence := NewTestStore(storage)
	encode := func(n ld.Node) (ld.Node, ld.Link) {
		lp := cidlink.LinkPrototype{Prefix: cid.Prefix{
			Version:  1,
			Codec:    0x71, // dag-cbor
			MhType:   0x12, // sha2-256
			MhLength: 4,
		}}

		lnk, err := persistence.Store(ld.LinkContext{}, lp, n)
		if err != nil {
			panic(err)
		}
		return n, lnk
	}

	var (
		leafAlpha, leafAlphaLnk         = encode(basicnode.NewString("alpha"))
		leafAlphaBlock, _               = blocks.NewBlockWithCid(storage[leafAlphaLnk], leafAlphaLnk.(cidlink.Link).Cid)
		leafBeta, leafBetaLnk           = encode(basicnode.NewString("beta"))
		leafBetaBlock, _                = blocks.NewBlockWithCid(storage[leafBetaLnk], leafBetaLnk.(cidlink.Link).Cid)
		middleMapNode, middleMapNodeLnk = encode(fluent.MustBuildMap(basicnode.Prototype.Map, 3, func(na fluent.MapAssembler) {
			na.AssembleEntry("foo").AssignBool(true)
			na.AssembleEntry("bar").AssignBool(false)
			na.AssembleEntry("nested").CreateMap(2, func(na fluent.MapAssembler) {
				na.AssembleEntry("alink").AssignLink(leafAlphaLnk)
				na.AssembleEntry("nonlink").AssignString("zoo")
			})
		}))
		middleMapBlock, _                 = blocks.NewBlockWithCid(storage[middleMapNodeLnk], middleMapNodeLnk.(cidlink.Link).Cid)
		middleListNode, middleListNodeLnk = encode(fluent.MustBuildList(basicnode.Prototype.List, 4, func(na fluent.ListAssembler) {
			na.AssembleValue().AssignLink(leafAlphaLnk)
			na.AssembleValue().AssignLink(leafAlphaLnk)
			na.AssembleValue().AssignLink(leafBetaLnk)
			na.AssembleValue().AssignLink(leafAlphaLnk)
		}))
		middleListBlock, _    = blocks.NewBlockWithCid(storage[middleListNodeLnk], middleListNodeLnk.(cidlink.Link).Cid)
		rootNode, rootNodeLnk = encode(fluent.MustBuildMap(basicnode.Prototype.Map, 4, func(na fluent.MapAssembler) {
			na.AssembleEntry("plain").AssignString("olde string")
			na.AssembleEntry("linkedString").AssignLink(leafAlphaLnk)
			na.AssembleEntry("linkedMap").AssignLink(middleMapNodeLnk)
			na.AssembleEntry("linkedList").AssignLink(middleListNodeLnk)
		}))
		rootBlock, _ = blocks.NewBlockWithCid(storage[rootNodeLnk], rootNodeLnk.(cidlink.Link).Cid)
	)
	return TestLDTree{
		Storage:           storage,
		LeafAlpha:         leafAlpha,
		LeafAlphaLnk:      leafAlphaLnk,
		LeafAlphaBlock:    leafAlphaBlock,
		LeafBeta:          leafBeta,
		LeafBetaLnk:       leafBetaLnk,
		LeafBetaBlock:     leafBetaBlock,
		MiddleMapNode:     middleMapNode,
		MiddleMapNodeLnk:  middleMapNodeLnk,
		MiddleMapBlock:    middleMapBlock,
		MiddleListNode:    middleListNode,
		MiddleListNodeLnk: middleListNodeLnk,
		MiddleListBlock:   middleListBlock,
		RootNode:          rootNode,
		RootNodeLnk:       rootNodeLnk,
		RootBlock:         rootBlock,
	}
}
